"""
This module contain methods to control arguments of functions
"""

__author__ = "Anna Igolkina"
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Anna Igolkina"
__email__ = "igolkinaanna11@gmail.com"

__all__ = ["weights_dbl",
           "weights_int",
           "probs",
           "individual",
           "lengths",
           "m_rounding",
           "n_solutions",
           "outcomes_sum"]

import numbers
import collections


def weights_dbl(weights):
    """
    Control whether weights parameter is a List of numbers
    """
    if not isinstance(weights, collections.Iterable):
        raise ValueError('Weights are not Iterable')
    if not all([isinstance(w, numbers.Number) for w in weights]):
        raise ValueError('Weights are not numbers')


def weights_int(weights):
    """
    Control whether weights parameter is a List of integer numbers
    """
    weights_dbl(weights)
    if not all([isinstance(w, int) for w in weights]):
        raise ValueError('Weights are not integer')


def probs(probs):
    """
    Control whether probs parameter is a List of real numbers in [0, 1]
    """
    if not isinstance(probs, collections.Iterable):
        raise ValueError('Probabilities are not Iterable')
    if not all([isinstance(p, numbers.Number) for p in probs]):
        raise ValueError('Probabilities are not numbers')
    if not all([0 <= p <= 1 for p in probs]):
        raise ValueError('Probabilities are not in [0, 1]')


def individual(indiv):
    """
    Control whether an individual is a list of 0 or 1
    """
    if not isinstance(indiv, collections.Iterable):
        raise ValueError('Individual is not Iterable')
    if not all([isinstance(t, int) for t in indiv]):
        raise ValueError('Target individual does not contain all numbers')
    if not all([t in (0, 1) for t in indiv]):
        raise ValueError('Target individual is a vector of [0 and 1]')


def lengths(*args):
    """
    Control whether arguments are lists of the same length
    """
    n = [len(element) for element in args]
    if len(n) == 0:
        raise ValueError('Please provide arguments')
    if not all([n[0] == elem for elem in n]):
        raise ValueError('Parameters should be the same length')


def m_rounding(m, weights):
    """
    Control the M-value for rounding
    """
    if not isinstance(m, int):
        raise ValueError('M for rounding should be an integer number')
    if m < len(weights):
        raise ValueError('M is lower than sum of weights')


def n_solutions(n):
    """
    Control the L-value for correction
    """
    if n is None:
        return
    if not isinstance(n, int):
        raise ValueError('N for linear programming should be an integer number')
    if n <= 0:
        raise ValueError('N for linear programming should be positive')


def outcomes_sum(outcomes, weights):
    """
    Control possible outcomes
    """
    error_message = 'Outcomes of a weighted sum are not correct'
    if isinstance(outcomes, collections.Iterable):
        if not all([isinstance(outcome, numbers.Number) for outcome in outcomes]):
            raise ValueError(error_message)
        if not all([0 <= outcome <= sum(weights) for outcome in outcomes]):
            raise ValueError(error_message)
    elif isinstance(outcomes, numbers.Number):
        if not (0 <= outcomes <= sum(weights)):
            raise ValueError(error_message)
    else:
        raise ValueError(error_message)



from .bernmix_control import *

__all__ = ["weights_dbl",
           "weights_int",
           "probs",
           "individual",
           "lengths",
           "m_rounding",
           "n_solutions",
           "outcomes_sum"]



"""
This module includes tools to correct CDF values for a weighted sum
of Bernoulli RVs using linear integer programming
"""

__author__ = "Anna Igolkina"
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Anna Igolkina"
__email__ = "igolkinaanna11@gmail.com"

__all__ = ["cdf_corrected"]

import numpy as np
from cvxopt import glpk, matrix

from . import bernmix_double as bmd
from .. import bernmix_int as bmi
from .. import bernmix_control as control


def cdf_corrected(probs, weights, target_indiv,
                  m_rounding=10**6, n_solutions=100):
    """
    This function computes the corrected value of CDF for a
    :param probs: a List of real numbers in the range [0,1]
                 representing probabilities of BRVs
    :param weights: a List of numbers; weights in a weighted sum of BRVs
    :param target_indiv: a List of binary outcomes of BRVs, 0/1 numbers
    :param m_rounding: a number of points to approximate
                       the weighted sum of BRVs
    :param n_solutions: a number of runs for linear integer programming
                        to correct the CDF value
    :return: an approximated and corrected CDF value for the target_indiv
    """

    # ----------------------------------------------
    # Control Input values
    # ----------------------------------------------
    control.weights_dbl(weights)
    control.probs(probs)
    control.individual(target_indiv)
    control.lengths(weights, probs, target_indiv)
    control.m_rounding(m_rounding, weights)
    control.n_solutions(n_solutions)
    # ----------------------------------------------

    n = len(weights)
    w = weights
    wc = bmd.weight_rounded(w, m_rounding)

    # S and Z values according to the article
    target_value_s = np.dot(w, target_indiv)
    target_value_z = np.dot(wc, target_indiv).astype(int)

    # number of points into two directions
    k = 10
    target_range = np.arange(target_value_z - k, target_value_z + k + 1, 1)
    distr_in_probs = np.zeros((2 * k + 1, 2))
    for i, target_value in enumerate(target_range):
        pop = binprog_multisol(wc, target_value, n_solutions, 10)
        if pop is None:
            distr_in_probs[i, 0] = 0
            distr_in_probs[i, 1] = 0
            continue

        s_values = list(map(lambda x: np.dot(x[0:n], w), pop))

        indiv_probs = list(map(lambda x: comp_indiv_prob(x[0:n], probs), pop))
        sum_prob = sum(indiv_probs)

        distr_in_probs[i, 0] = sum(np.extract(s_values <= target_value_s,
                                              indiv_probs)) / sum_prob
        distr_in_probs[i, 1] = sum(np.extract(s_values > target_value_s,
                                              indiv_probs)) / sum_prob

    pmf = bmi.pmf(probs, wc)

    pmf_range = list(map(lambda t: pmf[t], target_range))
    cdf2 = sum(pmf[:target_value_z + 1 - k - 1:]) + \
           np.dot(pmf_range, distr_in_probs[:, 0])

    return cdf2


def comp_indiv_prob(probs, indiv):
    """ Compute probability for an individual
    :param probs: a List of real numbers in the range [0,1]
                 representing probabilities of BRVs
    :param indiv: a List of binary outcomes of BRVs, 0/1 numbers
    """
    prob_multiply = list(map(lambda i, p: p if i == 1 else (1 - p),
                             indiv, probs))
    return np.prod(prob_multiply)


def binprog_multisol(weights, target_value, n_solutions, n_fixed=10):
    """
    This function returns multiple solutions of
    zero-one linear programming problem

    :param weights: a List of numbers; weights in a weighted sum of BRVs
    :param target_value: an outcome of the weighted sum of BRVs
    :param n_solutions: a number of runs for linear integer programming
                        to correct the CDF value
    :param n_fixed: an area around target_value utilised for correction
    :return: a
    """

    glpk.options['msg_lev'] = 'GLP_MSG_OFF'

    n = len(weights)
    c_init = np.append(weights, -target_value).astype(int)
    c = c_init.reshape((1, n+1))

    a_ident = np.identity(n + 1).astype(int)
    a_ident = np.delete(a_ident, n, axis=0)
    a_ub = np.concatenate((a_ident, -c), axis=0)
    b_ub = np.append(np.ones(n), [0]).astype(int).reshape(n+1, 1)

    pop = np.empty((0, n+1))
    for _ in range(n_solutions):
        # x_(n+1) = 1
        a_eq = np.append(np.zeros(n), [1]).astype(int).reshape(1, n+1)
        b_eq = np.asmatrix([1])

        idx = np.random.permutation(n)
        idx = idx[0:round(n_fixed/100 * n)]
        for i in idx:
            tmp = np.zeros(n+1)
            tmp[i] = 1
            a_eq = np.concatenate((a_eq, [tmp]), axis=0)
            b_eq = np.concatenate((b_eq, [np.random.randint(0, 2, 1)]), axis=0)

        c_mx = matrix(c, tc='d')
        g_mx = matrix(a_ub, tc='d')
        h_mx = matrix(b_ub, tc='d')
        a_mx = matrix(a_eq, tc='d')
        b_mx = matrix(b_eq, tc='d')

        status, x = glpk.ilp(c=c_mx,
                             G=g_mx,
                             h=h_mx,
                             A=a_mx,
                             b=b_mx,
                             B=set(range(n+1)))
        if status != 'optimal':
            continue

        pop = np.concatenate((pop, np.matrix(x).reshape((1, n+1))), axis=0)

    if pop.shape[1] == 0:
        return None

    pop = np.unique(pop, axis=0)
    # ANNA: from 0 to n
    return pop


if __name__ == "__main__":
    print('BernMix_correction is loaded')



import numbers
import numpy as np
import scipy.stats as stats
from scipy.stats import mvn
from scipy.optimize import minimize
from functools import partial
from random import seed
from itertools import product, combinations

def dp_int(p, w):
    """
    The dynamic algorithm to estimate PMF (probability mass function)
    :param p: vector of probabilities
    :param w: vector of weights
    :return: PMF
    """
    # Checks
    if len(p) != len(w):
        raise ValueError('Probabilities and weights should be of the same length')
    n = len(p)
    # print(n)
    if n == 0:
        raise ValueError('Probabilities and weights are empty')
    if not all([isinstance(w_tmp, int) for w_tmp in w]):
        raise ValueError('Weights are not integer')
    if not all([isinstance(p_tmp, numbers.Number) for p_tmp in p]):
        raise ValueError('Probabilities are not numbers')
    if not all([0 <= p_tmp <= 1 for p_tmp in p]):
        raise ValueError('Probabilities are not in [0, 1]')

    # Algorithm
    w_tot = sum(w)
    w_curr = 1
    s = [1] + [0] * w_tot  # you don't need + 1 here
    for i in range(n):
        w_curr = w_curr + w[i]
        # s_tmp = [0] * w[i] + s

        # print([s_tmp[2], s[2 - w[i]]])
        s[0:(w_curr)] = [s[j] * (1 - p[i]) + s[j- w[i]] * p[i]
                         if j >= w[i]
                         else s[j] * (1 - p[i])
                            for j in range(w_curr)]
    return s



def prob_stat(pa):
    """
    For a given probability distribution, this function calculates
    Mean, Variabce and all quantiles (including -Inf and +Inf)
    :param pa: PMF
    :return:
    """
    ma = sum([i * p for i, p in enumerate(pa)])
    va = sum([(i - ma) ** 2 * p for i, p in enumerate(pa)])
    pc_cum = [1 if p > 1 else p for p in np.cumsum(pa)]

    points = [-np.inf] + list(stats.norm.ppf(pc_cum, loc=0, scale=1))
    return ma, va, points


def discrepancy(pa, pb, v_ab, cov_ab):
    """
    Returns the discrepancy between true covariance value (v_ab)
     and covariance value from bvNormal approximation with (cov_ab)
    :param pa: PMF of one variable
    :param pb: PMF of another variable
    :param v_ab: True value of covariance
    :param cov_ab: Tested value of covariance
    :return: discrepancy
    """

    if abs(cov_ab) > 1: return np.inf
    ma, va, points_a = prob_stat(pa)

    mb, vb, points_b = prob_stat(pb)

    cov_mx = [[1, cov_ab], [cov_ab, 1]]

    joint_distr = np.empty((len(pa), len(pb)))
    for i in range(len(pa)):
        for j in range(len(pb)):
            res = mvn.mvnun([points_a[i], points_b[j]],
                            [points_a[i + 1], points_b[j + 1]], [0, 0], cov_mx)

            joint_distr[i, j] = res[0]


    cov_new = 0
    for i in range(len(pa)):
        for j in range(len(pb)):
            cov_new += joint_distr[i, j] * (i - ma) * (j - mb)

    return abs(cov_new - v_ab)



def joint_distr_cov(pa, pb, v_ab, echo=False):
    """
    Estimate table of joint distribution between two RVs
    :param pa: PMF of the first RV
    :param pb: PMF of the second RV
    :param v_ab: required covariance between RVs
    :return: table with joint distribution
    """

    if abs(v_ab) >= 1e-5:
        f_opt = partial(discrepancy, pa, pb, v_ab)
        # bnds = ((-1, 1), ())
        seed(239)
        res = minimize(f_opt, 0, method='SLSQP', bounds=[(-1, 1)],
                       options={'ftol': 1e-8, 'disp': False})
        cov_ab = res.x
        if echo:
            print(f'Optimised covariance is {cov_ab}')
        # print(f'Optimised covariance is {cov_ab}')

        if(np.isnan(cov_ab)):
            if v_ab > 0:
                cov_ab = 1
            else:
                cov_ab = -1
            if echo:
                print(f'New is {cov_ab}')
    else:
        cov_ab = 0

    ma, va, points_a = prob_stat(pa)
    mb, vb, points_b = prob_stat(pb)

    cov_mx = [[1, cov_ab], [cov_ab, 1]]


    joint_distr = np.empty((len(pa), len(pb)))
    for i in range(len(pa)):
        for j in range(len(pb)):
            res = mvn.mvnun([points_a[i], points_b[j]],
                            [points_a[i + 1], points_b[j + 1]], [0, 0], cov_mx)

            joint_distr[i, j] = res[0]
            if(np.isnan(joint_distr[i, j])):
                joint_distr[i, j] = 0
            if echo:
                print([i, j, res])

    cov_new = 0
    for i in range(len(pa)):
        for j in range(len(pb)):
            cov_new += joint_distr[i, j] * (i - ma) * (j - mb)

    if echo:
        print([v_ab, cov_new])
        print(f'Discrepancy: {sum(sum(joint_distr) - pb)}, {sum(joint_distr[:, 1] + joint_distr[:, 0] - pa)}, {v_ab- cov_new}')

    return joint_distr


def est_distr(p, w, v, v_zero=False):
    """
    Estimate PMF for a wighted sum of BRVs with known covariance matrix between RVs
    :param p: probabilities
    :param w: weights
    :param v: covariance matrix between BRVs
    :param v_zero: To set all covariance az zero
    :return:
    """
    # ---------------------------------------------------
    # Checks
    # ---------------------------------------------------
    if len(p) != len(w):
        raise ValueError('Probabilities and weights should be of the same length')
    n = len(p)
    if n == 0:
        raise ValueError('Probabilities and weights are empty')
    if not all([isinstance(w_tmp, int) for w_tmp in w]):
        raise ValueError('Weights are not integer')
    if not all([isinstance(p_tmp, numbers.Number) for p_tmp in p]):
        raise ValueError('Probabilities are not numbers')
    if not all([0 <= p_tmp <= 1 for p_tmp in p]):
        raise ValueError('Probabilities are not in [0, 1]')

    # ---------------------------------------------------
    #       Algorithm
    # ---------------------------------------------------

    w_tot = sum(w)
    w_curr = 1
    s = [1] + [0] * w_tot  # you don't need + 1 here
    for i in range(n):
        if i == 0:
            w_curr = w_curr + w[i]
            s_tmp = [0] * w[i] + s
            s[0:(w_curr)] = [s[j] * (1 - p[i]) + s_tmp[j] * p[i] for j in range(w_curr)]
        else:
            # break
            pa = s[:w_curr]
            # print(pa)
            v_ab = sum([w[j] * v[i, j] for j in range(i)])
            # print(v_ab)
            if v_zero:
                v_ab = 0

            if np.isnan(v_ab):
                raise ValueError('Variance is nan')
            pb = [1 - p[i], p[i]]

            # print(pa)
            # print(pb)
            joint_distr = joint_distr_cov(pa, pb, v_ab)
            # print(joint_distr)

            s_tmp1 = [0] * w[i] + list(joint_distr[:, 1])
            s_tmp2 = list(joint_distr[:, 0]) + [0] * w[i]
            w_curr = w_curr + w[i]
            s[0:(w_curr)] = [x1 + x2 for x1, x2 in zip(s_tmp1, s_tmp2)]

            # print(s[0:(w_curr)])
    return(s)


def get_v_max(pa, pb):
    """
    Function to generate max value of variance
    :param pa:
    :param pb:
    :return:
    """

    f_opt = lambda cov_ab: (-1) * discrepancy(pa, pb, 0, cov_ab)

    res = minimize(f_opt, 0, method='SLSQP',
                   options={'ftol': 1e-8, 'disp': False})
    v_ab = (-1) * res.fun

    j_distr = joint_distr_cov(pa, pb, v_ab, echo=False)

    return v_ab, j_distr


def dyn_tree(p, w):
    if len(p) != len(w):
        raise ValueError('Probabilities and weights should be of the same length')
    n = len(p)
    print(n)
    if n == 0:
        raise ValueError('Probabilities and weights are empty')
    if not all([isinstance(w_tmp, int) for w_tmp in w]):
        raise ValueError('Weights are not integer')
    if not all([isinstance(p_tmp, numbers.Number) for p_tmp in p]):
        raise ValueError('Probabilities are not numbers')
    if not all([0 <= p_tmp <= 1 for p_tmp in p]):
        raise ValueError('Probabilities are not in [0, 1]')

    s_blocks = []
    for i in range(n):
        s_tmp = [1 - p[i]] + [0] * (w[i] - 1) + [p[i]]
        s_blocks += [s_tmp]

    while len(s_blocks) > 1:
        s_len = len(s_blocks)
        m = int((s_len - s_len % 2) / 2)
        s_blocks_new = []
        for j in range(m):
            i = j * 2
            # if i == s_len:
            #     s_blocks_new += [s_blocks[i]]
            #     break
            b1 = s_blocks[i]
            b2 = s_blocks[i + 1]
            b_new = [0] * (len(b1) + len(b2) - 1)
            for i1 in range(len(b1)):
                if b1[i1] == 0:
                    continue
                for i2 in range(len(b2)):
                    i_new = i1 + i2
                    b_new[i_new] += b1[i1] * b2[i2]
            s_blocks_new += [b_new]
        if s_len % 2 == 1:
            s_blocks_new += [s_blocks[s_len - 1]]
        s_blocks = s_blocks_new
    return s_blocks



def dyn_tree_cov(p, w, v):
    if len(p) != len(w):
        raise ValueError('Probabilities and weights should be of the same length')
    n = len(p)
    if n == 0:
        raise ValueError('Probabilities and weights are empty')
    if not all([isinstance(w_tmp, int) for w_tmp in w]):
        raise ValueError('Weights are not integer')
    if not all([isinstance(p_tmp, numbers.Number) for p_tmp in p]):
        raise ValueError('Probabilities are not numbers')
    if not all([0 <= p_tmp <= 1 for p_tmp in p]):
        raise ValueError('Probabilities are not in [0, 1]')

    s_blocks = []
    for i in range(n):
        s_tmp = [1 - p[i]] + [0] * (w[i] - 1) + [p[i]]
        s_blocks += [s_tmp]

    v_blocks = v.copy()
    for i, j in combinations(range(n), 2):
        v_blocks[i, j] *= w[i] * w[j]
        v_blocks[j, i] *= w[i] * w[j]

    while len(s_blocks) > 1:
        n = len(s_blocks)

        var_block = [i for _, i, _ in [prob_stat(s) for s in s_blocks]]

        v_norm = v_blocks.copy()
        for i, j in combinations(range(n), 2):
            v_norm[i, j] /= (var_block[i] * var_block[j]) ** (1 / 2)
            v_norm[j, i] /= (var_block[i] * var_block[j]) ** (1 / 2)


        # find the pair with the highest variance
        i = j = 0
        if v_norm.max() == 0:
            i = 0
            j = 1
        else:
            while i == j:
                v_norm[i, j] = 0
                k = np.argmax(v_norm)
                j = k % n
                i = round((k - j) / n)
        # print(i, j)
        # print([v_blocks[i, j], v_norm[i, j]])


        pa = s_blocks[i]
        pb = s_blocks[j]
        v_ab = v_blocks[i, j]

        # print(pa)
        # print(pb)
        j_distr = joint_distr_cov(pa, pb, v_ab)

        v_bb = 0

        b_new = [0] * (len(pa) + len(pb) - 1)
        for i1, i2 in product(range(len(pa)),
                                   range(len(pb))):
            b_new[i1 + i2] += j_distr[i1][i2]

        if i > j:
            s_blocks.pop(i)
            s_blocks.pop(j)
        else:
            s_blocks.pop(j)
            s_blocks.pop(i)
        s_blocks += [b_new]

        v_new = list(v_blocks[i] + v_blocks[j])
        v_blocks = np.append(v_blocks, [v_new], axis=0)
        v_blocks = np.append(v_blocks.transpose(), [v_new+ [0]], axis=0)
        v_blocks = np.delete(v_blocks, [i, j], axis=0)
        v_blocks = np.delete(v_blocks, [i, j], axis=1)

        # print(b_new)




    return s_blocks[0]


"""
This module contain tools to compute CDF for a weighted sum of BRVs
with non-integer weights using rounding
"""

__author__ = "Anna Igolkina"
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Anna Igolkina"
__email__ = "igolkinaanna11@gmail.com"

__all__ = ["weight_rounded", "cdf_rounded"]

import numpy as np

from .. import bernmix_int as bmi
from .. import bernmix_control as control


def weight_rounded(weights, m_points):
    """
    This function return the scaled and rounded  approximation of weights
    :param weights: a List of numbers; weights in a weighted sum of BRVs
    :param m_points: a number of points to approximate
                       the weighted sum of BRVs
    :return:
    """

    # ----------------------------------------------
    # Control Input values
    # ----------------------------------------------
    control.weights_dbl(weights)
    control.m_rounding(m_points, weights)
    # ----------------------------------------------

    n = len(weights)
    weight_sum = sum(weights)
    c = (m_points + n) / weight_sum
    return np.around(weights * c).astype(int)


def cdf_rounded(probs, weights, indiv, m_rounding=10**6):
    """
    This function calculates the CDF value for several individuals
    using fixed number of points to approximate
    :param probs: a List of real numbers in the range [0,1]
                  representing probabilities of BRVs
    :param weights: a List of numbers; weights in a weighted sum of BRVs
    :param target_indivs: a list of individuals where each is a list of binary
                          outcomes of BRVs, 0/1 numbers
    :param m_rounding: a number of points to approximate
                       the weighted sum of BRVs
    :return:
    """

    # ----------------------------------------------
    # Control Input values
    # ----------------------------------------------
    control.weights_dbl(weights)
    control.probs(probs)
    control.individual(indiv)
    control.lengths(weights, probs, indiv)
    control.m_rounding(m_rounding, weights)
    # ----------------------------------------------

    # Compute approximated PMF
    weights_new = weight_rounded(weights, m_rounding)
    pmf = bmi.pmf(probs, weights_new)

    value = np.dot(weights_new, indiv).astype(int)
    cdf_value = sum(pmf[:(value+1)])

    return cdf_value



"""
This module includes tools to compute PMFs and CDFs for weighted sum of Bernoulli RVs
"""

import numpy as np

from . import bernmix_control as control
from . import bernmix_int as bmi
from . import bernmix_double as bmd
from . import bernmix_fancy as bm_new


def get_summary(pmf_bm, sum_bias, outcomes = None):
    """
    Prepare output data for PMF
    :param pmf_bm:
    :param sum_bias:
    :param outcomes:
    :return:
    """
    if outcomes is None:
        values = [v + sum_bias for v in range(0, len(pmf_bm))]
        return pmf_bm, values
    else:
        return pmf_bm[outcomes - sum_bias], outcomes


def normalise_params(probs, weights):
    """
    This function normalises parameters of probabilities and weights
    as to
    (1) make weights positive
    (2) remove trivial terms, probability = 0
    (3) remove shifting terms, probability = 1
    :param probs: vector of probabilities
    :param weights: vector of weights
    :return: tuple of * new probabilities
                      * new weights
                      * bias
    """

    # ----------------------------------------------
    # Control Input values
    # ----------------------------------------------
    control.weights_dbl(weights)
    control.probs(probs)
    control.lengths(weights, probs)
    # ----------------------------------------------

    sum_bias = sum([w for w in weights if w < 0])

    # BRV with negative weight is transformed to BRV with opposite probability
    probs = [p if w > 0 else 1 - p
                      for p, w in zip(probs, weights)]
    weights = [abs(w) for w in weights]

    # Remain only significant terms:
    # if some weights or probabilities equal to zero - remove
    idx_significant = [False if (p == 0) | (w == 0) else True for p, w in zip(probs, weights)]


    probs = [p for p, i in zip(probs, idx_significant) if i]
    weights = [w for w, i in zip(weights, idx_significant) if i]

    # Remove constant RVs : BRVs with probabilities equal to 1
    # and change bias
    idx_const = [p == 1 for p in probs]
    sum_bias += sum([w for w, i in zip(weights, idx_const) if i])
    probs = [p for p, i in zip(probs, idx_const) if not i]
    weights = [w for w, i in zip(weights, idx_const) if not i]

    return list(probs), list(weights), sum_bias


# def pmf_int_vals(probs, weights):
#     """
#     This function returns the PMF of the weighted sum of BRVs
#     when weights are integer
#     :param probs:
#     :param weights:
#     :param outcomes:
#     :return: The PMF across all possible values
#     """
#
#     # ----------------------------------------------
#     # Control Input values
#     # ----------------------------------------------
#     control.weights_dbl(weights)
#     control.probs(probs)
#     control.lengths(weights, probs)
#     # ----------------------------------------------
#
#     # remove trivial terms
#     probs, weights, sum_bias = normalise_params(probs, weights)
#     pmf_bm = bmi.pmf(probs, weights)
#
#     values = list(range(0, len(pmf_bm)))
#     values = [v + sum_bias for v in values]
#
#     return pmf_bm, values



def pmf_int(probs, weights, outcomes=None):
    """
    This function returns the PMF of the weighted sum of BRVs
    when weights are integer
    :param probs:
    :param weights:
    :param outcomes:
    :return: The PMF across all possible values
    """

    # ----------------------------------------------
    # Control Input values
    # ----------------------------------------------
    control.weights_dbl(weights)
    control.probs(probs)
    control.lengths(weights, probs)
    # ----------------------------------------------

    # remove trivial terms
    probs, weights, sum_bias = normalise_params(probs, weights)
    pmf_bm = bmi.pmf(probs, weights)

    return get_summary(pmf_bm, sum_bias, outcomes)


def pmf_int_conv(probs, weights, outcomes=None):
    """
    This function returns the PMF of the weighted sum of BRVs
    when weights are integer
    :param probs:
    :param weights:
    :param outcomes:
    :return: The PMF across all possible values
    """

    # ----------------------------------------------
    # Control Input values
    # ----------------------------------------------
    control.weights_dbl(weights)
    control.probs(probs)
    control.lengths(weights, probs)
    # ----------------------------------------------

    # remove trivial terms
    probs, weights, sum_bias = normalise_params(probs, weights)
    pmf_bm = bm_new.dp_int(probs, weights)

    return get_summary(pmf_bm, sum_bias, outcomes)


def pmf_int_dep(probs, weights, cov):
    """
    This function returns the PMF of the weighted sum of BRVs
    when weights are integer
    :param probs: pro
    :param weights:
    :param cov: covariance matrix
    :return: The PMF across all possible values
    """

    # ----------------------------------------------
    # Control Input values
    # ----------------------------------------------
    control.weights_dbl(weights)
    control.probs(probs)
    control.lengths(weights, probs)
    # ----------------------------------------------

    # remove trivial terms
    probs, weights, sum_bias = normalise_params(probs, weights)
    pmf_bm = bm_new.dyn_tree_cov(probs, weights, cov)

    return get_summary(pmf_bm, sum_bias)

def cdf_int(probs, weights, outcomes=None):
    """
    This function returns the CDF of the weighted sum of BRVs
    when weights are integer
    when weights are integer
    :param probs:
    :param weights:
    :param outcomes:
    :return: The PMF across all possible values
    """

    # ----------------------------------------------
    # Control Input values
    # ----------------------------------------------
    control.probs(probs)
    control.weights_dbl(weights)
    control.lengths(weights, probs)
    # ----------------------------------------------

    # remove trivial terms
    probs, weights, sum_bias = normalise_params(probs, weights)
    # compute PMF
    pmf_bm = bmi.pmf(probs, weights)
    cdf_bm = np.cumsum(pmf_bm)

    return get_summary(cdf_bm, sum_bias, outcomes)


def cdf_double(probs, weights, target_indiv,
               m_rounding=10**6, n_solutions=None):
    """
    This function reputrn the vector of probabilities for possible values
    of the weighted sum of Bernoulli random variables
    when weights are double
    :param probs: a List of real numbers in the range [0,1]
                 representing probabilities of BRVs
    :param weights: a List of numbers; weights in a weighted sum of BRVs
    :param target_indiv: a List of binary outcomes of BRVs, 0/1 numbers
    :param m_rounding: a number of points to approximate
                       the weighted sum of BRVs
    :param n_solutions: a number of runs for linear integer programming
                        to correct the CDF value
    :return: an approximated or corrected CDF value for the target_indiv
    """

    # ----------------------------------------------
    # Control Input values
    # ----------------------------------------------
    control.weights_dbl(weights)
    control.probs(probs)
    control.individual(target_indiv)
    control.lengths(weights, probs, target_indiv)
    control.m_rounding(m_rounding, weights)
    control.n_solutions(n_solutions)
    # ----------------------------------------------

    if n_solutions is None:
        cdf_value = bmd.cdf_rounded(probs, weights, target_indiv, m_rounding)
    else:
        cdf_value = bmd.cdf_corrected(probs, weights, target_indiv,
                                  m_rounding, n_solutions)

    return cdf_value


def cdf_permut(probs, weights, target_indiv, n_permut=10 ** 6):
    """
    Get CDF py permutations/simulations
    :param probs: A list of real numbers in the range [0,1]
                 representing probabilities of BRVs
    :param weights: A list of numbers; weights in a weighted sum of BRVs
    :param target_indivs: A list of individual, where each is a list of binary
                          outcomes of BRVs, list with 0 or 1 numbers
    :param n_permut: Number of permutations
    :return: CDF value
    """

    # ----------------------------------------------------
    # Control Input values
    # ----------------------------------------------------
    control.weights_dbl(weights)
    control.probs(probs)
    control.individual(target_indiv)
    control.lengths(weights, probs, target_indiv)
    control.n_solutions(n_permut)
    # ----------------------------------------------------

    # generate a population of Multivariate BRV with target probabilities
    pop = np.zeros((len(probs), n_permut), int)  # Pre-allocate matrix
    for i, p in enumerate(probs):
        pop[i] = np.random.binomial(1, p, n_permut)
    pop = np.transpose(pop)

    # Compute outcomes of the weighted sum of BRVs for the population
    pop_values = list(map(lambda indiv: np.dot(weights, indiv), pop))

    # Compute outcomes of the weighted sum of BRVs for the target individuals
    target_value = np.dot(weights, target_indiv)

    # Compute the approximation of CDF for target_values
    cdf_value = sum(pop_values <= target_value) / n_permut

    return cdf_value


def pmf_int_bf(probs, weights):
    """
    This function caclulates pmf fpr each individual by brute-force search
    :param probs: A list of real numbers in the range [0,1]
                 representing probabilities of BRVs
    :param weights: A list of numbers; weights in a weighted sum of BRVs
    :return: pmf_bm - a vector with discrete values of PMF: P(x) = pmf_bm[x]
    """

    # ----------------------------------------------
    # Control Input values
    # ----------------------------------------------
    control.weights_dbl(weights)
    control.probs(probs)
    control.lengths(weights, probs)
    # ----------------------------------------------

    prob_indiv, outcomes = conv_all_outcomes(probs, weights)
    outcomes_unique = list(set(outcomes))
    outcomes_unique.sort()
    pmf_bm = [sum(prob_indiv[outcomes == i])
              for i in outcomes_unique]
    return pmf_bm, outcomes_unique


def conv_all_outcomes(probs, weights):
    """
    This function calculates probabilities for each outcome
    :param probs: vector of probabilities
    :param weights: vector of weights
    :return: probabilities and outcomes
    """

    def comp_indiv_prob(indiv, probs):
        """ Compute probability for an individual """
        prob_multiply = list(map(lambda ind, p: p if ind == 1 else (1 - p),
                                 indiv, probs))
        return np.prod(prob_multiply)

    # ----------------------------------------------
    # Control Input values
    # ----------------------------------------------
    control.weights_dbl(weights)
    control.probs(probs)
    control.lengths(weights, probs)
    # ----------------------------------------------

    n_terms = len(probs)
    # initialise size of outcomes
    outcomes = np.zeros(2 ** n_terms)
    prob_indiv = np.zeros(2 ** n_terms)

    # initialise two first of outcomes (0,0,...0) and (1,0,...0)
    outcomes[0:2] = [0, weights[0]]
    prob_indiv[0] = comp_indiv_prob(np.zeros(n_terms), probs)
    prob_indiv[1] = comp_indiv_prob(np.append([1], np.zeros(n_terms-1)), probs)

    for i in range(1, n_terms):
        n = 2 ** i
        outcomes[n:2 * n] = outcomes[0:n] + weights[i]
        prob_indiv[n:2 * n] = prob_indiv[0:n] / (1 - probs[i]) * probs[i]

    return prob_indiv, outcomes

#
# def poibinmix_pmf_int(probs, wights):
#     pass
#
#
# def poibinmix_cdf_double(probs, wights, target_value, n_points = None):
#     pass
#
#
# def binmix_pmf_int(probs, num_of_trails, wights):
#     pass
#
#
# def binmix_cdf_double(probs, num_of_trails, wights, target_value,
#                       n_points = None):
#     pass
#
#
# def radmix_pmf_int(probs, wights):
#     pass
#


if __name__ == "__main__":
    pass



cimport bernmix_fourier
from libc.stdlib cimport malloc, free

def pmf(probs, weights):
    # ANNA
    cdef int n = len(weights)
    cdef int sw = 0
    for wi in weights:
        sw += wi

    cdef int *w_new = <int *>malloc(n * sizeof(int))
    cdef double *p_new = <double *>malloc(n * sizeof(double))
    for i in range(n):
        w_new[i] = weights[i]
        p_new[i] = probs[i]

    cdef double *pmf_res = <double *> malloc((sw + 1) * sizeof(double))
    bernmix_fourier.pmf_bern_mixture(n, p_new, w_new, pmf_res)

    pmf_result = [ pmf_res[i] for i in range(sw+1) ]
    free(pmf_res)
    return pmf_result






"""
This module aims to
"""

import numpy as np
from bm_dynam import *
from bernmix.bernmix_int import pmf as bmi

def bm_joint(p, w_vec, v = None, tol = 1e-5):
    """
    Function to predict joint distribution of
    :param p: pribabilities of Bernoulli RVs
    :param w_vec: vector of vectors of weights for RVs
    :param tol: tolerance
    :return:
    """
    # Checks
    # TODO

    # -----------------------------------
    # Preparation
    # -----------------------------------

    # Algorithm


    w_max = [sum(w) + 1 for w in w_vec]
    m_max = np.cumprod(w_max)

    w_new = w_vec[0]
    for i in range(1, len(w_vec)):
        w_new = [w1 + w2 * m_max[i-1] for w1, w2 in zip(w_new, w_vec[i])]

    if v is None:
        s = bmi(p, w_new)
    else:
        s = dyn_tree_cov(p, w_new, v)

    res = [[i, val] for i, val in enumerate(s) if val > tol]
    prob_sum = sum([val for _, val in res])
    res = [[i, val / prob_sum] for i, val in res]

    i = 7598


    res_brv = []
    for i, val in res:
        y = []
        for j in range(len(w_max)):
            y_tmp = i % w_max[j]
            y += [y_tmp]
            i -= y_tmp
            i /= w_max[j]
            i = round(i)
        res_brv += [y + [val]]







# setup.py
from setuptools import setup, find_packages
from distutils.extension import Extension
from Cython.Build import build_ext


PACKAGE = "bernmix"
NAME = "bernmix"
DESCRIPTION = "Methods to compute PMF and CDF values of a weighted sum of " \
              "i.ni.d. BRVs"
AUTHOR = "Anna Igolkina"
AUTHOR_EMAIL = "igolkinaanna11@gmail.com"
URL = "https://github.com/iganna/bernmix"
VERSION = "1.11.31"

#with open('requirements.txt') as f:
#    reqs = f.read().splitlines()
reqs = ['cvxopt', 'Cython', 'numpy', 'scipy']

with open("README.md", "r") as fh:
    long_description = fh.read()


src_dir = 'bernmix/bernmix_int'
ext = Extension(src_dir + '.bernmix_int',
                [src_dir + '/bernmix_int.pyx',
                 src_dir + '/bernmix_fourier.c'],
                libraries=['fftw3'],
                extra_compile_args=["-lfftw3"])
package_data = {'bernmix_int': ['*.pxd', '*.pyx', '*.c', '*.h']}

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license="MIT",
    url=URL,
    packages=find_packages(),
    install_requires=reqs,
    ext_modules=[ext],
    cmdclass={'build_ext': build_ext},
    package_data = package_data,
    include_package_data=True,
    long_description=long_description,
    long_description_content_type="text/markdown", 
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    zip_safe=False,
)



from .bernmix import *

__author__ = "Anna Igolkina"
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Anna Igolkina"
__email__ = "igolkinaanna11@gmail.com"
__status__ = "Development"

__all__ = ["pmf_int",
           "cdf_int",
           "cdf_double",
           "cdf_permut",
           "normalise_params",
           # "conv_all_outcomes",
           "pmf_int_dep",
           "pmf_int_bf",
           "pmf_int_conv"
           ]
